import logging
import shlex
import subprocess
import sys
from pathlib import Path
from typing import Any, override

import attrs
import git

from liblaf import grapes
from liblaf.cherries import core

logger: logging.Logger = logging.getLogger(__name__)


@attrs.define
class Git(core.PluginSchema):
    commit: bool = False
    inputs: list[Path] = attrs.field(factory=list)
    outputs: list[Path] = attrs.field(factory=list)
    repo: git.Repo = attrs.field(default=None)
    temps: list[Path] = attrs.field(factory=list)
    verify: bool = False

    @override
    @core.impl(before=("Comet",))
    def end(self, *args, **kwargs) -> None:
        if self.commit and self.repo.is_dirty(untracked_files=True):
            try:
                self.repo.git.add(all=True)
                subprocess.run(["git", "status"], check=False)
                message: str = self._make_commit_message()
                self.repo.git.commit(message=message, no_verify=not self.verify)
            except git.GitCommandError:
                logger.exception("")
        self.run.log_other("cherries.git.sha", self.repo.head.commit.hexsha)

    @override
    @core.impl
    def log_input(self, path: Path, name: Path, **kwargs) -> None:
        self.inputs.append(self._relative_to_repo(path))

    @override
    @core.impl
    def log_output(self, path: Path, name: Path, **kwargs) -> None:
        self.outputs.append(self._relative_to_repo(path))

    @override
    @core.impl
    def log_temp(self, path: Path, name: Path, **kwargs) -> None:
        self.temps.append(self._relative_to_repo(path))

    @override
    @core.impl
    def start(self, *args, **kwargs) -> None:
        self.repo = git.Repo(search_parent_directories=True)

    def _make_commit_message(self) -> str:
        name: str = self.run.exp_name
        message: str = f"chore(cherries): {name}\n\n"
        meta: dict[str, Any] = {}
        if url := self.run.url:
            meta["url"] = url
        meta["exp_dir"] = self.run.exp_dir.relative_to(self.repo.working_dir)
        meta["cwd"] = Path.cwd().relative_to(self.repo.working_dir)
        meta["cmd"] = shlex.join(sys.orig_argv)
        if params := self.run.get_params():
            meta["params"] = params
        if inputs := self.inputs:
            meta["inputs"] = inputs
        if outputs := self.outputs:
            meta["outputs"] = outputs
        if temps := self.temps:
            meta["temps"] = temps
        message += grapes.yaml.encode(meta).decode()
        return message

    def _relative_to_repo(self, path: Path) -> Path:
        if self.repo is None:
            return path
        try:
            return path.relative_to(self.repo.working_dir)
        except ValueError:
            return path



from ._impl import ImplInfo, collect_impls, get_impl_info, impl
from ._methods import (
    asset,
    end,
    get_other,
    get_others,
    get_param,
    get_params,
    get_url,
    input,  # noqa: A004
    log_asset,
    log_input,
    log_metric,
    log_metrics,
    log_other,
    log_others,
    log_output,
    log_param,
    log_params,
    log_temp,
    output,
    run,
    set_step,
    start,
    temp,
)
from ._plugin import Plugin
from ._plugin_manager import PluginManager
from ._plugin_schema import PluginSchema
from ._run import Run
from ._typing import MethodName, PluginId

__all__ = [
    "ImplInfo",
    "MethodName",
    "Plugin",
    "PluginId",
    "PluginManager",
    "PluginSchema",
    "Run",
    "asset",
    "collect_impls",
    "end",
    "get_impl_info",
    "get_other",
    "get_others",
    "get_param",
    "get_params",
    "get_url",
    "impl",
    "input",
    "log_asset",
    "log_input",
    "log_metric",
    "log_metrics",
    "log_other",
    "log_others",
    "log_output",
    "log_param",
    "log_params",
    "log_temp",
    "output",
    "run",
    "set_step",
    "start",
    "temp",
]



from typing import override

from liblaf.cherries import core, plugins

from ._abc import Profile


class ProfileDefault(Profile):
    @override  # impl Profile
    def init(self) -> core.Run:
        run: core.Run = core.run
        run.register(plugins.Comet(disabled=False))
        run.register(plugins.Git(commit=True))
        run.register(plugins.Local())
        run.register(plugins.Logging())
        return run



from liblaf.cherries import core, profiles
from liblaf.cherries.profiles import ProfileLike


def start(profile: ProfileLike | None = None) -> core.Run:
    profile = profiles.factory(profile)
    run: core.Run = profile.init()
    run.start()
    return run



from . import bundle, config, core, meta, plugins
from ._main import end, main, start
from ._version import __version__, __version_tuple__
from .config import BaseConfig
from .core import (
    Run,
    asset,
    get_other,
    get_others,
    get_param,
    get_params,
    input,  # noqa: A004
    log_asset,
    log_input,
    log_metric,
    log_metrics,
    log_other,
    log_others,
    log_output,
    log_param,
    log_params,
    log_temp,
    output,
    run,
    set_step,
    temp,
)

__all__ = [
    "BaseConfig",
    "Run",
    "__version__",
    "__version_tuple__",
    "asset",
    "bundle",
    "config",
    "core",
    "end",
    "get_other",
    "get_others",
    "get_param",
    "get_params",
    "input",
    "log_asset",
    "log_input",
    "log_metric",
    "log_metrics",
    "log_other",
    "log_others",
    "log_output",
    "log_param",
    "log_params",
    "log_temp",
    "main",
    "meta",
    "output",
    "plugins",
    "run",
    "set_step",
    "start",
    "temp",
]



from typing import Protocol

import giturlparse as _giturlparse


class GitUrlParsed(Protocol):
    """.

    References:
        1. <https://github.com/nephila/giturlparse/blob/master/README.rst>
    """

    @property
    def host(self) -> str: ...
    @property
    def platform(self) -> str: ...
    @property
    def owner(self) -> str: ...
    @property
    def repo(self) -> str: ...


def giturlparse(url: str) -> GitUrlParsed:
    info: GitUrlParsed = _giturlparse.parse(url)  # pyright: ignore[reportAssignmentType]
    return info



from typing import Literal

from environs import env

from ._abc import Profile

# ensure profiles are registered
from ._default import ProfileDefault  # noqa: F401
from ._playground import ProfilePlayground  # noqa: F401

type ProfileName = Literal["default", "playground"] | str  # noqa: PYI051
type ProfileLike = ProfileName | Profile | type[Profile]


def factory(profile: ProfileLike | None = None) -> Profile:
    if profile is None and env.bool("DEBUG", False):
        profile = "playground"
    if profile is None:
        profile = env.str("PROFILE", "default")
    if isinstance(profile, str):
        return Profile[profile]()
    if isinstance(profile, Profile):
        return profile
    return profile()



import logging
from collections.abc import Mapping
from pathlib import Path
from typing import Any, override

import attrs
from liblaf.grapes.logging import autolog

from liblaf import grapes
from liblaf.cherries import core


@attrs.define
class Logging(core.PluginSchema):
    @property
    def log_file(self) -> Path:
        return self.run.logs_dir / self.run.entrypoint.with_suffix(".log").name

    @override
    @core.impl(before=("Comet",))
    def end(self, *args, **kwargs) -> None:
        self.run.log_asset(self.log_file)

    @override
    @core.impl
    def log_metric(
        self, name: str, value: Any, step: int | None = None, **kwargs
    ) -> None:
        __tracebackhide__ = True
        if step is None:
            autolog.info("%s: %s", name, value)
        else:
            autolog.info("step: %s, %s: %s", step, name, value)

    @override
    @core.impl
    def log_metrics(
        self, metrics: Mapping[str, Any], step: int | None = None, **kwargs
    ) -> None:
        __tracebackhide__ = True
        if step is None:
            autolog.info("%s", metrics)
        else:
            autolog.info("step: %s, %s", step, metrics)

    @override
    @core.impl
    def start(self, *args, **kwargs) -> None:
        grapes.logging.init(file=self.log_file, force=True)
        logging.getLogger("urllib3.connectionpool").setLevel(logging.CRITICAL)



from collections.abc import Mapping
from pathlib import Path
from typing import Any

from _typeshed import StrPath

from ._run import Run

run: Run

def asset(path: StrPath, *, mkdir: bool = False) -> Path: ...
def input(path: StrPath, *, mkdir: bool = False) -> Path: ...  # noqa: A001
def output(path: StrPath, *, mkdir: bool = False) -> Path: ...
def temp(path: StrPath, *, mkdir: bool = False) -> Path: ...
def end(*args, **kwargs) -> None: ...
def get_other(name: str) -> Any: ...
def get_others() -> Mapping[str, Any]: ...
def get_param(name: str) -> Any: ...
def get_params() -> Mapping[str, Any]: ...
def get_url() -> str: ...
def log_asset(path: StrPath, **kwargs) -> None: ...
def log_input(path: StrPath, **kwargs) -> None: ...
def log_output(path: StrPath, **kwargs) -> None: ...
def log_temp(path: StrPath, **kwargs) -> None: ...
def log_metric(name: str, value: Any, step: int | None = None, **kwargs) -> None: ...
def log_metrics(
    metrics: Mapping[str, Any], step: int | None = None, **kwargs
) -> None: ...
def log_other(name: str, value: Any) -> None: ...
def log_others(others: Mapping[str, Any]) -> None: ...
def log_param(name: str, value: Any) -> None: ...
def log_params(params: Mapping[str, Any]) -> None: ...
def set_step(step: int | None = None) -> None: ...
def start(*args, **kwargs) -> None: ...



from ._abc import Bundle, BundleItem
from ._landmarks import BundleLandmarks
from ._registry import BundleRegistry, bundles
from ._series import BundleSeries
from ._utils import relative_to_or_name

__all__ = [
    "Bundle",
    "BundleItem",
    "BundleLandmarks",
    "BundleRegistry",
    "BundleSeries",
    "bundles",
    "relative_to_or_name",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import attrs

from ._typing import PluginId

if TYPE_CHECKING:
    from ._plugin_manager import PluginManager


def _default_id(self: Plugin) -> PluginId:
    return type(self).__name__


@attrs.define
class Plugin:
    id: PluginId = attrs.field(
        default=attrs.Factory(_default_id, takes_self=True), kw_only=True
    )
    manager: PluginManager = attrs.field(
        default=None, repr=False, init=False, kw_only=True
    )



from __future__ import annotations

from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING, Literal, override

import attrs
import pydantic

from ._abc import Bundle, BundleItem

if TYPE_CHECKING:
    from _typeshed import StrPath


def snake_to_kebab(snake: str) -> str:
    return snake.replace("_", "-")


class File(pydantic.BaseModel):
    name: str
    time: float


class Series(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(alias_generator=snake_to_kebab)
    file_series_version: Literal["1.0"] = "1.0"
    files: list[File] = []


@attrs.define
class BundleSeries(Bundle):
    @override
    def match(self, path: Path) -> bool:
        return path.suffix == ".series"

    @override
    def ls_files(self, path: Path, prefix: Path) -> Generator[BundleItem]:
        series: Series = Series.model_validate_json(path.read_bytes())
        for meta in series.files:
            absolute: Path = path.parent / meta.name
            relative: StrPath
            try:
                relative = absolute.relative_to(prefix)
            except ValueError:
                relative = meta.name
            yield BundleItem(absolute, relative, required=True)



import logging
import unittest.mock
from collections.abc import Mapping
from pathlib import Path
from typing import Any, override

import attrs
import comet_ml
import cytoolz as toolz
import dvc.api
import dvc.exceptions
import git

from liblaf import grapes
from liblaf.cherries import core, meta

logger: logging.Logger = logging.getLogger(__name__)


@attrs.frozen
class Asset:
    path: Path
    name: Path
    metadata: Mapping[str, Any] | None = None
    kwargs: Mapping[str, Any] = attrs.field(factory=dict)


@attrs.define
class Comet(core.PluginSchema):
    disabled: bool = attrs.field(default=False)
    _assets_git: list[Asset] = attrs.field(factory=list)

    @override
    @core.impl(after=("Git", "Logging"))
    def end(self, *args, **kwargs) -> None:
        try:
            self._log_asset_git_end()
        except git.GitError:
            logger.exception("")
        self.experiment.end()

    @override
    @core.impl
    def get_other(self, name: str) -> Any:
        return self.experiment.get_other(name)

    @override
    @core.impl
    def get_others(self) -> Mapping[str, Any]:
        return self.experiment.others

    @override
    @core.impl
    def get_param(self, name: str) -> Any:
        return self.experiment.get_parameter(name)

    @override
    @core.impl
    def get_params(self) -> Mapping[str, Any]:
        return self.experiment.params

    @override
    @core.impl
    def get_url(self) -> str:
        return self.experiment.url  # pyright: ignore[reportReturnType]

    @override
    @core.impl
    def log_asset(self, path: Path, name: Path, **kwargs) -> None:
        if self._log_asset_git(path, name, **kwargs):
            return
        self.experiment.log_asset(path, name.as_posix(), **kwargs)

    @override
    @core.impl
    def log_input(
        self,
        path: Path,
        name: Path,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        name = "inputs" / name
        metadata = toolz.assoc(metadata or {}, "type", "input")
        self.log_asset(path, name, metadata=metadata, **kwargs)

    @override
    @core.impl
    def log_metric(
        self, name: str, value: Any, step: int | None = None, **kwargs
    ) -> None:
        return self.experiment.log_metric(name, value, step=step, **kwargs)

    @override
    @core.impl
    def log_metrics(
        self, metrics: Mapping[str, Any], step: int | None = None, **kwargs
    ) -> None:
        return self.experiment.log_metrics(dict(metrics), step=step, **kwargs)

    @override
    @core.impl
    def log_other(self, name: str, value: Any) -> None:
        return self.experiment.log_other(name, value)

    @override
    @core.impl
    def log_others(self, others: Mapping[str, Any]) -> None:
        return self.experiment.log_others(dict(others))

    @override
    @core.impl
    def log_output(
        self,
        path: Path,
        name: Path,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        name = "outputs" / name
        metadata = toolz.assoc(metadata or {}, "type", "output")
        self.log_asset(path, name, metadata=metadata, **kwargs)

    @override
    @core.impl
    def log_param(self, name: str, value: Any) -> None:
        return self.experiment.log_parameter(name, value)

    @override
    @core.impl
    def log_params(self, params: Mapping[str, Any]) -> None:
        return self.experiment.log_parameters(dict(params))

    @override
    @core.impl(after=("Logging",))
    def start(self, *args, **kwargs) -> None:
        try:
            comet_ml.start(
                project_name=self.run.project_name,
                experiment_config=comet_ml.ExperimentConfig(
                    disabled=self.disabled, name=self.run.exp_name
                ),
            )
        except ValueError:
            logger.exception("")

    @override
    @core.impl
    def set_step(self, step: int | None = None) -> None:
        return self.experiment.set_step(step)

    @property
    def experiment(self) -> comet_ml.CometExperiment:
        return comet_ml.get_running_experiment() or unittest.mock.MagicMock()

    def _log_asset_dvc(
        self,
        path: Path,
        name: Path,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> bool:
        try:
            # ? I don't know why, but `dvc.api.get_url` only works with this. Maybe a DVC bug?
            dvc_path: Path = path.absolute().relative_to(Path.cwd())
            uri: str = dvc.api.get_url(str(dvc_path))
        except dvc.exceptions.OutputNotFoundError:
            return False
        dvc_file: Path = path.with_name(path.name + ".dvc")
        dvc_meta: Mapping[str, Any] = grapes.yaml.load(dvc_file)
        metadata: dict[str, Mapping] = toolz.merge(metadata or {}, dvc_meta["outs"][0])
        self.experiment.log_remote_asset(
            uri, name.as_posix(), metadata=metadata, **kwargs
        )
        return True

    def _log_asset_git(
        self,
        path: Path,
        name: Path,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> bool:
        try:
            repo = git.Repo(search_parent_directories=True)
        except git.InvalidGitRepositoryError:
            return False
        try:
            if repo.ignored(path):
                return False
        except git.GitCommandError:
            # `path` may be outside repository
            return False
        self._assets_git.append(
            Asset(path=path, name=name, metadata=metadata, kwargs=kwargs)
        )
        return True

    def _log_asset_git_end(self) -> None:
        if len(self._assets_git) == 0:
            return
        repo = git.Repo(search_parent_directories=True)
        info: meta.GitUrlParsed = meta.giturlparse(repo.remote().url)
        for asset in self._assets_git:
            uri: str
            match str(info.platform):
                case "github":
                    assert repo.working_tree_dir is not None
                    absolute: Path = Path(asset.path).absolute()
                    relative: str = absolute.relative_to(
                        repo.working_tree_dir
                    ).as_posix()
                    sha: str = repo.head.commit.hexsha
                    uri = f"https://{info.host}/{info.owner}/{info.repo}/raw/{sha}/{relative}"
                case _:
                    uri = asset.path.as_posix()
            self.experiment.log_remote_asset(
                uri,
                asset.name.as_posix(),
                metadata=dict(asset.metadata) if asset.metadata is not None else None,
                **asset.kwargs,
            )



from __future__ import annotations

import functools
import graphlib
import logging
import math
from collections.abc import Callable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, overload

import attrs
import cachetools
import wrapt

from ._impl import ImplInfo, collect_impls, get_impl_info
from ._typing import MethodName, PluginId

if TYPE_CHECKING:
    from ._plugin import Plugin


logger: logging.Logger = logging.getLogger(__name__)


@overload
def delegate[C: Callable](func: C, /, *, first_result: bool = False) -> C: ...
@overload
def delegate[C: Callable](*, first_result: bool = False) -> Callable[[C], C]: ...
def delegate(func: Callable | None = None, *, first_result: bool = False) -> Any:
    if func is None:
        return functools.partial(delegate, first_result=first_result)

    @wrapt.decorator
    def wrapper(
        wrapped: Callable,
        instance: PluginManager,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        __tracebackhide__ = True
        return instance.delegate(
            method=wrapped.__name__, args=args, kwargs=kwargs, first_result=first_result
        )

    return wrapper(func)


@attrs.define
class PluginManager:
    plugins: dict[PluginId, Plugin] = attrs.field(factory=dict, kw_only=True)

    def register(self, plugin: Plugin) -> None:
        for method_name in collect_impls(plugin):
            self._sort_plugins_cache.pop(method_name, None)
        plugin.manager = self
        self.plugins[plugin.id] = plugin

    def delegate(
        self,
        method: MethodName,
        args: Sequence[Any] = (),
        kwargs: Mapping[str, Any] = {},
        *,
        first_result: bool = False,
    ) -> Any:
        __tracebackhide__ = True
        results: list[Any] = []
        for plugin in self._sort_plugins(method):
            try:
                result: Any = getattr(plugin, method)(*args, **kwargs)
            except Exception:
                logger.exception("Plugin %s", plugin.id)
            else:
                if result is None:
                    continue
                if first_result:
                    return result
                results.append(result)
        if first_result:
            return None
        return results

    _sort_plugins_cache: cachetools.Cache[MethodName, Sequence[Plugin]] = attrs.field(
        repr=False, init=False, factory=lambda: cachetools.Cache(math.inf)
    )

    def _sort_plugins_key(self, method_name: MethodName) -> MethodName:
        return method_name

    @cachetools.cachedmethod(
        lambda self: self._sort_plugins_cache, key=_sort_plugins_key
    )
    def _sort_plugins(self, method_name: MethodName) -> Sequence[Plugin]:
        plugins: dict[PluginId, Plugin] = {}
        sorter: graphlib.TopologicalSorter[PluginId] = graphlib.TopologicalSorter()
        for plugin in self.plugins.values():
            method: Callable | None = getattr(plugin, method_name, None)
            impl_info: ImplInfo | None = get_impl_info(method)
            if impl_info is None:
                continue
            plugins[plugin.id] = plugin
            sorter.add(plugin.id, *impl_info.after)
            for before_id in impl_info.before:
                sorter.add(before_id, plugin.id)
        plugins_sorted: list[Plugin] = []
        for plugin_id in sorter.static_order():
            plugin: Plugin | None = plugins.get(plugin_id)
            if plugin is None:
                continue
            plugins_sorted.append(plugin)
        return plugins_sorted



from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any

import attrs

from ._plugin import Plugin

if TYPE_CHECKING:
    from ._run import Run


@attrs.define
class PluginSchema(Plugin):
    @property
    def run(self) -> Run:
        return self.manager  # pyright: ignore[reportReturnType]

    def end(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def get_other(self, name: str) -> Any:
        raise NotImplementedError

    def get_others(self) -> Mapping[str, Any]:
        raise NotImplementedError

    def get_param(self, name: str) -> Any:
        raise NotImplementedError

    def get_params(self) -> Mapping[str, Any]:
        raise NotImplementedError

    def get_url(self) -> str:
        raise NotImplementedError

    def log_asset(self, path: Path, name: Path, **kwargs) -> None:
        raise NotImplementedError

    def log_input(self, path: Path, name: Path, **kwargs) -> None:
        raise NotImplementedError

    def log_metric(
        self, name: str, value: Any, step: int | None = None, **kwargs
    ) -> None:
        raise NotImplementedError

    def log_metrics(
        self, metrics: Mapping[str, Any], step: int | None = None, **kwargs
    ) -> None:
        raise NotImplementedError

    def log_other(self, name: str, value: Any) -> None:
        raise NotImplementedError

    def log_others(self, others: Mapping[str, Any]) -> None:
        raise NotImplementedError

    def log_output(self, path: Path, name: Path, **kwargs) -> None:
        raise NotImplementedError

    def log_param(self, name: str, value: Any) -> None:
        raise NotImplementedError

    def log_params(self, params: Mapping[str, Any]) -> None:
        raise NotImplementedError

    def log_temp(self, path: Path, name: Path, **kwargs) -> None:
        raise NotImplementedError

    def set_step(self, step: int | None = None) -> None:
        raise NotImplementedError

    def start(self, *args, **kwargs) -> None:
        raise NotImplementedError



import inspect
import itertools
import typing
from collections.abc import Callable, Mapping, Sequence
from inspect import Parameter
from typing import Any

import pydantic

from liblaf.cherries import core, profiles

from ._start import start


def main[T](
    main: Callable[..., T], *, profile: profiles.ProfileLike | None = None
) -> T:
    run: core.Run = start(profile=profile)
    args: Sequence[Any]
    kwargs: Mapping[str, Any]
    args, kwargs = _make_args(main)
    configs: list[pydantic.BaseModel] = [
        arg
        for arg in itertools.chain(args, *kwargs.values())
        if isinstance(arg, pydantic.BaseModel)
    ]
    for config in configs:
        run.log_params(config.model_dump(mode="json"))
    try:
        return main(*args, **kwargs)
    finally:
        run.end()


def _make_args(func: Callable) -> tuple[Sequence[Any], Mapping[str, Any]]:
    hints: dict[str, Any] = typing.get_type_hints(func)
    signature: inspect.Signature = inspect.signature(func)
    args: list[Any] = []
    kwargs: dict[str, Any] = {}
    for name, param in signature.parameters.items():
        annotation: Any = hints.get(name)
        match param.kind:
            case Parameter.POSITIONAL_ONLY:
                args.append(_make_arg(param, annotation))
            case Parameter.POSITIONAL_OR_KEYWORD | Parameter.KEYWORD_ONLY:
                kwargs[name] = _make_arg(param, annotation)
            case _:
                pass
    return args, kwargs


def _make_arg(param: Parameter, annotation: Any) -> Any:
    if param.default is not Parameter.empty:
        return param.default
    if param.annotation is not Parameter.empty:
        if not isinstance(param.annotation, str):
            return param.annotation()
        if annotation is not None:
            return annotation()
    return None



from ._abc import Profile
from ._default import ProfileDefault
from ._factory import ProfileLike, ProfileName, factory
from ._playground import ProfilePlayground

__all__ = [
    "Profile",
    "ProfileDefault",
    "ProfileLike",
    "ProfileName",
    "ProfilePlayground",
    "factory",
]



from collections.abc import Generator
from pathlib import Path

import attrs

from ._abc import Bundle, BundleItem


def _default_registry() -> list[Bundle]:
    from ._landmarks import BundleLandmarks
    from ._series import BundleSeries

    return [BundleLandmarks(), BundleSeries()]


@attrs.define
class BundleRegistry:
    registry: list[Bundle] = attrs.field(factory=_default_registry)

    def ls_files(self, path: Path, prefix: Path) -> Generator[BundleItem]:
        for bundle in self.registry:
            if bundle.match(path):
                yield from bundle.ls_files(path, prefix)

    def register(self, bundle: Bundle) -> None:
        self.registry.append(bundle)


bundles: BundleRegistry = BundleRegistry()



from collections.abc import Generator
from pathlib import Path
from typing import override

import attrs

from ._abc import Bundle, BundleItem
from ._utils import relative_to_or_name


@attrs.define
class BundleLandmarks(Bundle):
    suffixes: set[str] = attrs.field(
        factory=lambda: {
            ".obj",
            ".ply",
            ".stl",
            ".vti",
            ".vtk",
            ".vtp",
            ".vtr",
            ".vts",
            ".vtu",
        }
    )

    @override
    def match(self, path: Path) -> bool:
        return path.suffix in self.suffixes

    @override
    def ls_files(self, path: Path, prefix: Path) -> Generator[BundleItem]:
        absolute: Path = path.with_suffix(".landmarks.json")
        yield BundleItem(
            absolute, relative_to_or_name(absolute, prefix), required=False
        )



import logging
import shutil
from pathlib import Path
from typing import override

import attrs
from liblaf.grapes.logging import RichFileHandler, as_filter, autolog

from liblaf.cherries import core

logger: logging.Logger = logging.getLogger(__name__)


@attrs.define
class Local(core.PluginSchema):
    folder: Path = attrs.field(default=None)

    @property
    def log_file(self) -> Path:
        return self.folder / "logs" / self.run.entrypoint.with_suffix(".log").name

    @override
    @core.impl
    def log_asset(self, path: Path, name: Path, **kwargs) -> None:
        __tracebackhide__ = True
        target: Path = self.folder / name
        self._copy(path, target)

    @override
    @core.impl
    def log_input(self, path: Path, name: Path, **kwargs) -> None:
        __tracebackhide__ = True
        target: Path = self.folder / "inputs" / name
        self._copy(path, target)

    @override
    @core.impl
    def log_output(self, path: Path, name: Path, **kwargs) -> None:
        __tracebackhide__ = True
        target: Path = self.folder / "outputs" / name
        self._copy(path, target)

    @override
    @core.impl
    def log_temp(self, path: Path, name: Path, **kwargs) -> None:
        __tracebackhide__ = True
        target: Path = self.folder / "temp" / name
        self._copy(path, target)

    @override
    @core.impl(after=("Logging",))
    def start(self, *args, **kwargs) -> None:
        local_dir: Path = self.run.exp_dir / ".cherries"
        local_dir.mkdir(parents=True, exist_ok=True)
        (local_dir / ".gitignore").write_text("*\n")
        entrypoint: Path = self.run.entrypoint
        self.folder = (
            local_dir
            / entrypoint.stem
            / self.run.start_time.strftime("%Y-%m-%dT%H%M%S")
        )
        self._copy(entrypoint, self.folder / "src" / entrypoint.name)
        self._config_logging()

    def _config_logging(self) -> None:
        logger: logging.Logger = logging.getLogger()
        handler = RichFileHandler(self.log_file)
        handler.addFilter(as_filter())
        logger.addHandler(handler)

    def _copy(self, source: Path, target: Path) -> None:
        __tracebackhide__ = True
        if target.exists():
            if target.samefile(self.log_file):
                return
            autolog.warning("Overwriting existing file: %s", target)
        target.parent.mkdir(parents=True, exist_ok=True)
        if source.is_dir():
            shutil.copytree(source, target, dirs_exist_ok=True)
        else:
            shutil.copy2(source, target)



from __future__ import annotations

import functools
import logging
import sys
from collections.abc import Mapping
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

import attrs
import git
import git.exc
from liblaf.grapes.logging import autolog

from liblaf.cherries.bundle import bundles, relative_to_or_name
from liblaf.cherries.core._typing import MethodName

from ._plugin_manager import PluginManager, delegate

if TYPE_CHECKING:
    from _typeshed import StrPath

logger: logging.Logger = logging.getLogger(__name__)

_PATH_SKIP_NAMES: set[str] = {"exp", "src"}


@attrs.define
class Run(PluginManager):
    _assets_queue: list[Path] = attrs.field(init=False, factory=list)
    _inputs_queue: list[Path] = attrs.field(init=False, factory=list)
    _outputs_queue: list[Path] = attrs.field(init=False, factory=list)
    _temps_queue: list[Path] = attrs.field(init=False, factory=list)

    @functools.cached_property
    def data_dir(self) -> Path:
        return self.exp_dir / "data"

    @functools.cached_property
    def entrypoint(self) -> Path:
        if sys.argv[0] == "-c":
            return None  # pyright: ignore[reportReturnType]
        return Path(sys.argv[0]).resolve()

    @functools.cached_property
    def exp_dir(self) -> Path:
        parent: Path = self.entrypoint.parent
        while parent.name in _PATH_SKIP_NAMES:
            parent = parent.parent
        return parent

    @functools.cached_property
    def exp_name(self) -> str:
        name: str = self.entrypoint.relative_to(self.project_dir).as_posix()
        while True:
            original: str = name
            for folder in _PATH_SKIP_NAMES:
                name = name.removeprefix(f"{folder}/")
            if name == original:
                break
        return name

    @functools.cached_property
    def fig_dir(self) -> Path:
        return self.exp_dir / "fig"

    @functools.cached_property
    def logs_dir(self) -> Path:
        return self.exp_dir / "logs"

    @functools.cached_property
    def project_dir(self) -> Path:
        try:
            repo: git.Repo = git.Repo(search_parent_directories=True)
        except git.exc.InvalidGitRepositoryError as err:
            logger.warning("%s", err)
            return Path.cwd().resolve()
        else:
            return Path(repo.working_dir).resolve()

    @functools.cached_property
    def project_name(self) -> str:
        return self.project_dir.name

    @functools.cached_property
    def start_time(self) -> datetime:
        return datetime.now().astimezone()

    @functools.cached_property
    def temp_dir(self) -> Path:
        return self.exp_dir / "temp"

    @property
    def url(self) -> str:
        return self.get_url()

    def asset(self, path: StrPath, *, mkdir: bool = False) -> Path:
        absolute: Path = self.exp_dir / path
        if mkdir:
            absolute.parent.mkdir(parents=True, exist_ok=True)
        self._assets_queue.append(absolute)
        return absolute

    def input(self, path: StrPath, *, mkdir: bool = False) -> Path:
        absolute: Path = self.data_dir / path
        if mkdir:
            absolute.parent.mkdir(parents=True, exist_ok=True)
        self._inputs_queue.append(absolute)
        return absolute

    def output(self, path: StrPath, *, mkdir: bool = False) -> Path:
        absolute: Path = self.data_dir / path
        if mkdir:
            absolute.parent.mkdir(parents=True, exist_ok=True)
        self._outputs_queue.append(absolute)
        return absolute

    def temp(self, path: StrPath, *, mkdir: bool = False) -> Path:
        absolute: Path = self.temp_dir / path
        if mkdir:
            absolute.parent.mkdir(parents=True, exist_ok=True)
        self._temps_queue.append(absolute)
        return absolute

    def end(self, *args, **kwargs) -> None:
        __tracebackhide__ = True
        self.log_other("cherries.end_time", datetime.now().astimezone())
        for path in self._assets_queue:
            self.log_asset(path)
        for path in self._inputs_queue:
            self.log_input(path)
        for path in self._outputs_queue:
            self.log_output(path)
        for path in self._temps_queue:
            self.log_temp(path)
        self.delegate("end", args, kwargs)

    @delegate(first_result=True)
    def get_other(self, name: str) -> Any: ...

    @delegate(first_result=True)
    def get_others(self) -> Mapping[str, Any]: ...

    @delegate(first_result=True)
    def get_param(self, name: str) -> Any: ...

    @delegate(first_result=True)
    def get_params(self) -> Mapping[str, Any]: ...

    @delegate(first_result=True)
    def get_url(self) -> str: ...

    def log_asset(self, path: StrPath, **kwargs) -> None:
        __tracebackhide__ = True
        self._log_asset(path, "log_asset", self.exp_dir, **kwargs)

    def log_input(self, path: StrPath, **kwargs) -> None:
        __tracebackhide__ = True
        self._log_asset(path, "log_input", self.data_dir, **kwargs)

    def log_output(self, path: StrPath, **kwargs) -> None:
        __tracebackhide__ = True
        self._log_asset(path, "log_output", self.data_dir, **kwargs)

    def log_temp(self, path: StrPath, **kwargs) -> None:
        __tracebackhide__ = True
        self._log_asset(path, "log_temp", self.temp_dir, **kwargs)

    @delegate
    def log_metric(
        self, name: str, value: Any, step: int | None = None, **kwargs
    ) -> None: ...

    @delegate
    def log_metrics(
        self, metrics: Mapping[str, Any], step: int | None = None, **kwargs
    ) -> None: ...

    @delegate
    def log_other(self, name: str, value: Any) -> None: ...

    @delegate
    def log_others(self, others: Mapping[str, Any]) -> None: ...

    @delegate
    def log_param(self, name: str, value: Any) -> None: ...

    @delegate
    def log_params(self, params: Mapping[str, Any]) -> None: ...

    @delegate
    def set_step(self, step: int | None = None) -> None: ...

    def start(self, *args, **kwargs) -> None:
        __tracebackhide__ = True
        self.delegate("start", args, kwargs)
        self.log_other(
            "cherries.entrypoint",
            _relative_or_absolute(self.entrypoint, self.project_dir),
        )
        self.log_other(
            "cherries.exp_dir", _relative_or_absolute(self.exp_dir, self.project_dir)
        )
        self.log_other("cherries.start_time", self.start_time)

    def _log_asset(
        self, path: StrPath, method_name: MethodName, prefix: StrPath, **kwargs
    ) -> None:
        __tracebackhide__ = True
        path: Path = Path(path).resolve()
        if not path.exists():
            autolog.warning("No such file or directory: %s", path)
            return
        prefix: Path = Path(prefix).resolve()
        name: StrPath = relative_to_or_name(path, prefix)
        self.delegate(method_name, args=(path, name), kwargs=kwargs)
        for absolute, relative, required in bundles.ls_files(path, prefix):
            absolute = Path(absolute)  # noqa: PLW2901
            relative = Path(relative)  # noqa: PLW2901
            if not absolute.exists():
                if required:
                    autolog.warning("No such file or directory: %s", absolute)
                continue
            self.delegate(method_name, args=(absolute, relative), kwargs=kwargs)


def _relative_or_absolute(path: Path, prefix: Path) -> Path:
    try:
        return path.relative_to(prefix)
    except ValueError:
        return path



from __future__ import annotations

import abc
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

if TYPE_CHECKING:
    from _typeshed import StrPath


class BundleItem(NamedTuple):
    path: StrPath
    relative: StrPath
    required: bool


class Bundle(abc.ABC):
    @abc.abstractmethod
    def match(self, path: Path) -> bool: ...

    @abc.abstractmethod
    def ls_files(self, path: Path, prefix: Path) -> Iterable[BundleItem]:
        raise NotImplementedError



import functools
import inspect
from collections.abc import Callable, Iterable
from typing import Any, overload

import attrs
import wrapt

from liblaf import grapes

from ._typing import MethodName, PluginId


@attrs.define
class ImplInfo:
    after: Iterable[PluginId] = ()
    before: Iterable[PluginId] = ()


@overload
def impl[C: Callable](
    func: C, /, *, after: Iterable[PluginId] = (), before: Iterable[PluginId] = ()
) -> C: ...
@overload
def impl[C: Callable](
    *, after: Iterable[PluginId] = (), before: Iterable[PluginId] = ()
) -> Callable[[C], C]: ...
def impl[**P, T](func: Callable[P, T] | None = None, /, **kwargs) -> Any:
    if func is None:
        return functools.partial(impl, **kwargs)

    info = ImplInfo(**kwargs)

    @wrapt.decorator
    def wrapper(
        wrapped: Callable[P, T],
        _instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> T:
        __tracebackhide__ = True
        return wrapped(*args, **kwargs)

    func = wrapper(func)
    grapes.wrapt_setattr(func, "impl", info)
    return func


def collect_impls(cls: Any) -> dict[MethodName, ImplInfo]:
    if not isinstance(cls, type):
        cls = type(cls)
    impls: dict[MethodName, ImplInfo] = {}
    for name, method in inspect.getmembers(cls):
        info: ImplInfo | None = get_impl_info(method)
        if info is not None:
            impls[name] = info
    return impls


def get_impl_info(func: Callable | None) -> ImplInfo | None:
    if func is None:
        return None
    return grapes.wrapt_getattr(func, "impl", None)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import StrPath


def relative_to_or_name(path: Path, prefix: Path) -> StrPath:
    try:
        return path.relative_to(prefix)
    except ValueError:
        return path.name
